/* $Id: bodc2harmgen.c 2736 2007-10-20 16:59:04Z flaterco $ */

/* Translate BODC format to harmgen format.  See usage in main program. */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <time.h>

#define linelen 1000

/* Lose trailing space */
void strip (char *buf) {
  int i;
  char c;
  assert (buf);
  i = strlen(buf);
  while (i) {
    c = buf[--i];
    if (!(c == '\n' || c == '\r' || c == ' ' || c == '\t'))
      break;
    buf[i] = '\0';
  }
}

/*************************************************************************
The following code for managing timestamps is recycled from Timestamp.cc
in XTide version 2.6.3.  Some modifications have been made.
*************************************************************************/

#define compare_int(a,b) (((int)(a))-((int)(b)))
int
compare_tm (struct tm a, struct tm b) {
  int temp;
  if ((temp = compare_int (a.tm_year, b.tm_year)))
    return temp;
  if ((temp = compare_int (a.tm_mon, b.tm_mon)))
    return temp;
  if ((temp = compare_int (a.tm_mday, b.tm_mday)))
    return temp;
  if ((temp = compare_int (a.tm_hour, b.tm_hour)))
    return temp;
  if ((temp = compare_int (a.tm_min, b.tm_min)))
    return temp;
  return compare_int (a.tm_sec, b.tm_sec);
}

time_t tm2utc (struct tm ht) {
  time_t guess = 0;
  int loopcounter = sizeof(time_t) * 8;
  time_t thebit = ((time_t)1) << (loopcounter-1);

  // Are we signed?
  if (thebit < (time_t)0) {
    switch (loopcounter) {
    case 32:
      // Get this sign thing out of the way.
      if (ht.tm_year + 1900 < 1970)
	guess |= 0x80000000;
      loopcounter = 31;
      break;
    case 64:
      // Had a lot of trouble with overflows.  40 bits should suffice for
      // all reasonable requests.
      if (ht.tm_year + 1900 < 1970)
	guess |= 0xFFFFFF0000000000LL;
      loopcounter = 40;
      break;
    default:
      // Wha?!?
      // Skip the sign bit and hope for the best.
      // You can't just shift thebit right because it propagates the sign bit.
      loopcounter--;
    }
    thebit = ((time_t)1) << (loopcounter-1);
    assert (thebit > (time_t)0);
  } else {
    // We are unsigned.
    // Assuming no overflows, no action required here.
    ;
  }

  int compare;
  struct tm* gm;
  for (; loopcounter; loopcounter--) {
    time_t newguess = guess | thebit;
    // If gmtime fails, we assume that we are too far in the future.
    gm = gmtime(&newguess);
    if (!(compare = (gm ? compare_tm (*gm, ht) : 1)))
      return newguess;
    if (compare < 0)
      guess = newguess;
    assert (thebit > (time_t)0);
    thebit >>= 1;
  }

  assert (!thebit);
  gm = gmtime(&guess);
  assert (!(gm ? compare_tm (*gm, ht) : 1));
  return guess;
}

/*************************************************************************/

void do_bodc_header (char *fname) {
  char buf[linelen];
  char station_id[linelen], station_name[linelen];
  double lat, lng;
  FILE *fp, *fpout;

  if (!(fp = fopen (fname, "r"))) {
    perror (fname);
    exit (-1);
  }

  if (!(fpout = fopen ("header.txt", "w"))) {
    perror ("header.txt");
    exit (-1);
  }

  /* Read Proudman header */

  /* Port: */
  assert (fgets (buf, linelen, fp));
  assert (sscanf (buf, "Port: %s", station_id) == 1);

  /* Site: */
  assert (fgets (buf, linelen, fp));
  assert (!strncmp (buf, "Site:", 5));
  assert (strlen (buf) > 20);
  strcpy (station_name, buf+19);
  strip (station_name);

  /* Latitude: */
  assert (fgets (buf, linelen, fp));
  assert (sscanf (buf, "Latitude: %lf", &lat) == 1);

  /* Longitude: */
  assert (fgets (buf, linelen, fp));
  assert (sscanf (buf, "Longitude: %lf", &lng) == 1);

  /* Start Date: */
  /* End Date: */
  /* We process multiple files so these would be incorrect anyway. */
  /* Also, some files start over in the middle. */
  assert (fgets (buf, linelen, fp));
  assert (fgets (buf, linelen, fp));

  /* Contributor: */
  /* Hard-coded */
  assert (fgets (buf, linelen, fp));
  assert (strlen (buf) > 20);
  assert (!strncmp (buf+19, "Proudman Oceanographic Laboratory", 33));

  /* Datum information: */
  /* Hard-coded */
  assert (fgets (buf, linelen, fp));
  assert (strstr (buf, "Admiralty Chart Datum"));

  /* Parameter code: */
  /* These can change, so skip it. */
  assert (fgets (buf, linelen, fp));
#if 0
  assert (strlen (buf) > 20);
  strcpy (parameter_code, buf+19);
  strip (parameter_code);
#endif

  /* Sanity check */
  assert (fgets (buf, linelen, fp));
  assert (!strncmp (buf, "  Cycle", 7));

  /* Produce harmgen header file */

  fprintf (fpout, "%s, UK\n", station_name);
  fprintf (fpout, ":Europe/London\nmeters\n");

  fprintf (fpout, "\
# The harmonic constants used to perform tide predictions for locations\n\
# in the U.K. are derived from sea level data obtained from the British\n\
# Oceanographic Data Centre (BODC) based at the Proudman Oceanographic\n\
# Laboratory, Liverpool.  The predictions are not the same as those\n\
# computed by the Proudman Oceanographic Laboratory, which uses its own\n\
# sets of harmonic constants.\n\
#\n\
# The data were supplied by the British Oceanographic Data Centre as\n\
# part of the function of the National Tidal & Sea Level Facility,\n\
# hosted by the Proudman Oceanographic Laboratory and funded by the\n\
# Department for Environment, Food & Rural Affairs and the Natural\n\
# Environment Research Council.\n\
#\n\
# For more details please read http://flaterco.com/pol.html.\n\
# BEGIN HOT COMMENTS\n\
# country: U.K.\n\
# source: Derived from BODC data with Harmgen 2.2\n\
# restriction: Non-commercial use only\n\
# station_id_context: POL\n\
# station_id: %s\n\
# datumkind: Admiralty Chart Datum\n\
# legalese: READ flaterco.com/pol.html\n\
# !units: meters\n\
# !longitude: %.4f\n\
# !latitude: %.4f\n", station_id, lng, lat);

  fclose (fp);
  fclose (fpout);
}

void do_bodc_data (char *fname, FILE *rawfile) {
  char buf[linelen];
  FILE *fp, *qa=NULL;
  long int linecount=0, parsecount=0, usedcount=0;
  printf ("%s ", fname);
  if (!(fp = fopen (fname, "r"))) {
    perror (fname);
    exit (-1);
  }
  while (fgets (buf, linelen, fp)) {
    struct tm t;
    int cycle;
    double amp, res;
    char flag;
    linecount++;
    if (sscanf (buf, "%d) %d/%d/%d %d:%d:%d %lf%c %lf",
      &cycle, &(t.tm_year), &(t.tm_mon), &(t.tm_mday), &(t.tm_hour),
		&(t.tm_min), &(t.tm_sec), &amp, &flag, &res) == 10) {
      parsecount++;
      if (flag == ' ') {
	if (!(res > -4 && res < 4 && amp > -5 && amp < 20)) {
          if (!qa) {
	    if (!(qa = fopen ("QA.txt", "a"))) {
	      perror ("QA.txt");
	      exit (-1);
	    }
	    fprintf (qa, "Rejected records from %s\n", fname);
          }
	  fprintf (qa, "%s", buf);
          printf ("!");
	} else {
	  usedcount++;
	  t.tm_year -= 1900;
	  t.tm_mon -= 1;
	  fprintf (rawfile, "%lld %f\n", (long long int)(tm2utc(t)), amp);
	}
      } else {
        printf ("%c", flag);
      }
    }
  }
  fclose (fp);
  if (qa)
    fclose (qa);
  printf (" %ld lines, %ld parsed, %ld used\n", linecount, parsecount, usedcount);
}

int main (int argc, char **argv) {
  int i;

  if (argc < 2) {
    fprintf (stderr, "\
  Usage:  bodc2harmgen bodcfile.txt [... more files ...]\n\
  E.g.:  bodc2harmgen 2000ABE.txt 2001ABE.txt 2002ABE.txt\n\
\n\
  Output:  harmgen_input.txt    (raw format input for harmgen)\n\
           header.txt           (the header file for harmgen)\n\
  Any flags found in the input are dumped to standard out as a kind of\n\
  progress indicator.  Flagged data are not used.\n");
    exit (-1);
  }

  FILE *fp;
  do_bodc_header (argv[1]);

  if (!(fp = fopen ("harmgen_input.txt", "w"))) {
    perror ("harmgen_input.txt");
    exit (-1);
  }

  for (i=1; i<argc; i++)
    do_bodc_data (argv[i], fp);

  exit (0);
}
