/* $Id: bsh2harmgen.cc 2736 2007-10-20 16:59:04Z flaterco $ */

/* Translate BSH (German) format to Harmgen format. */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <time.h>

#include "Dstr.hh"

// Year:  the year (for 1900 < year < 2100)
// Yday:  day in the year, from 1 to 366.
// Hour:  hours and fractional hours
// Returns:  time_t for default time zone (no DST)
time_t weird_time_format_to_time_t (int year, int yday, double hour) {
  struct tm tm;
  assert (year > 1900 && year < 2100);
  tm.tm_year = year - 1900;
  //               J   F   M   A   M   J   J   A   S   O   N   D
  int days[12] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
  if (!(year % 4)) // No good before 1901 or after 2099
    days[1]++;

  tm.tm_mon = 0;
  tm.tm_mday = yday;
  while (tm.tm_mday > days[tm.tm_mon]) {
    tm.tm_mday -= days[tm.tm_mon];
    assert (tm.tm_mon < 12);
    tm.tm_mon++;
  }
  tm.tm_hour = (int)hour;
  hour -= tm.tm_hour;
  hour *= 60;
  tm.tm_min = (int)hour;
  hour -= tm.tm_min;
  hour *= 60;
  tm.tm_sec = (int)hour;

  tm.tm_isdst = 0;
  time_t ret = mktime(&tm);
  assert (ret != -1);
  return ret;
}

struct gs {
  unsigned station_id;
  double latdatum;
  char *name;
  char *realname;
  char *nlat;
  char *elng;
};

gs metadata[12] = {
  {103,2.7, "Bremerhaven AL",     "Bremerhaven, Alter Leuchtturm","53 32 42","8 34 05"},
  {502,2.54,"Bremen Oslebshausen","Bremen, Oslebshausen",         "53 07 11","8 42 43"}, 
  {504,2.14,"Brunsbuettel",       "Brunsbttel",                  "53 53 16","9 07 33"},
  {505,2.33,"Buesum",             "Bsum",                        "54 07 11","8 51 30"},
  {101,1.9, "Borkum",             "Borkum, Fischerbalje",         "53 33 27","6 44 52"},
  {506,2.23,"Cuxhaven",           "Cuxhaven, Steubenhft",        "53 52 03","8 43 06"},
  {507,2.49,"Emden",              "Emden, Neue Seeschleuse",      "53 20 13","7 11 11"},
  {508,2.29,"Hamburg",            "Hamburg, St. Pauli",           "53 32 44","9 58 12"},
  {509,1.89,"Helgoland",          "Helgoland",                    "54 10 44","7 53 24"},
  {510,2.56,"Husum",              "Husum",                        "54 28 20","9 01 29"},
  {111,1.88,"Norderney",          "Norderney, Riffgat",           "53 41 48","7 09 28"},
  {512,2.76,"Wilhelmshaven",      "Wilhelmshaven, Alter Vorhafen","53 30 52","8 08 42"}
};

int main(int argc, char **argv) {
  if (argc != 2) {
    fprintf (stderr, "\
  Usage:  bsh2harmgen bshfile.txt\n\
\n\
  Output:  harmgen_input.txt    (raw format input for harmgen)\n\
           header.txt           (the header file for harmgen)\n");
    exit (-1);
  }

  assert (setenv ("TZ", "LST-1", 1) == 0);
  tzset ();
  assert (timezone == -3600);

  FILE *fp, *headout, *hgout;
  if (!(fp = fopen (argv[1], "r"))) {
    perror (argv[1]);
    exit (-1);
  }
  if (!(headout = fopen ("header.txt", "w"))) {
    perror ("header.txt");
    exit (-1);
  }
  if (!(hgout = fopen ("harmgen_input.txt", "w"))) {
    perror ("harmgen_input.txt");
    exit (-1);
  }

  Dstr buf;
  int gotbe = 0, gotnam=0, index=0;
  unsigned byr, bmo, bday, eyr, emo, eday, junk;
  while (1) {
    buf.getline(fp);
    assert (!buf.isNull());
    {
      double hr, height;
      int year, yday;
      if (sscanf (buf.aschar(), "%u %u %d %d %lf %lf",
		  &junk, &junk, &year, &yday, &hr, &height) == 6)
        break;
    }
    buf.trim();
    if (!gotnam) {
      for (index=0; index<12; index++) {
        if (buf == metadata[index].name) {
          gotnam = 1;
          break;
        }
      }
      assert (gotnam);
      fprintf (headout, "%s, Germany\n", metadata[index].realname);
      fprintf (headout, ":Europe/Berlin\nmeters\n");
    }
    if (buf != "") {
      fprintf (headout, "# %s\n", buf.aschar());
      if (!gotbe) {
        if (sscanf (buf.aschar(), "from %u %u %u %u - %u %u %u", &byr, &bmo,
		    &bday, &junk, &eyr, &emo, &eday) == 7)
          gotbe = 1;
      }
    }
  }

  fprintf (headout, "#\n# BEGIN HOT COMMENTS\n");
  fprintf (headout, "# rel_pedigree: 500000000 (Harmgen)\n# Datum: Lowest Astronomical Tide\n# Source: Derived from BSH data with Harmgen 2.2\n# !units: meters\n");
  double deg, min, sec, lat, lng;
  assert (sscanf (metadata[index].nlat, "%lf %lf %lf", &deg, &min, &sec) == 3);
  lat = deg + min/60 + sec/3600;
  assert (sscanf (metadata[index].elng, "%lf %lf %lf", &deg, &min, &sec) == 3);
  lng = deg + min/60 + sec/3600;
  fprintf (headout, "# !longitude: %lf\n# !latitude: %lf\n", lng, lat);
  fprintf (headout, "# DATUMOVERRIDE: %lf\n", metadata[index].latdatum);
  fprintf (headout, "# station_id_context: BSH\n# station_id: %u\n",
	   metadata[index].station_id);
  fclose (headout);

  while (!buf.isNull()) {
    double hr, height;
    int year, yday;
    assert (sscanf (buf.aschar(), "%u %u %d %d %lf %lf",
      &junk, &junk, &year, &yday, &hr, &height) == 6);
    height /= 100;
    time_t t = weird_time_format_to_time_t (year, yday, hr);
#ifdef VERBOSE
    char tbuf[80];
    strftime (tbuf, 80, "%Y-%m-%d %H:%M:%S %Z", localtime(&t));
    printf ("%s  %lf\n", tbuf, height);
#endif
    fprintf (hgout, "%ld %lf\n", t, height);
    buf.getline(fp);
  }
  fclose (hgout);
  fclose (fp);
  exit (0);
}
