#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <time.h>

#include "Dstr.hh"

// csv2harm < something.csv
// Produces harmgen_input.txt and header.txt

// Remove the field name from input, verifying that it is the field expected.
// Also get rid of trailing ^M
void stripfield (Dstr &buf, const char *fieldname) {
  assert (fieldname);
  buf.trim_tail();
  assert (buf.strstr (fieldname) == 0);
  int i = buf.strchr (',');
  assert (i == (int)strlen(fieldname));
  buf /= i+1;
}


/*

http://www.timetemperature.com/tzca/canada_time_zone.shtml

Canada uses six primary time zones. From east to west they are
Newfoundland Time Zone, Atlantic Time Zone, Eastern Time, Central Time
Zone, Mountain Time Zone, and the Pacific Time Zone.

Some areas of Canada not using Daylight Saving Time include, Fort
St. John, Charlie Lake, Taylor and Dawson Creek in British Columbia,
Creston in the East Kootenays, and most of Saskatchewan (except Denare
Beach and Creighton).

Maps at http://inms-ienm.nrc-cnrc.gc.ca/en/time_services/daylight_saving_e.php

From http://inms-ienm.nrc-cnrc.gc.ca/en/faq_time_e.php

1 The areas of Qubec, east of 63 west longitude, do not change to
daylight time and remain on Atlantic Standard time all year round.

2 The areas of Ontario, west of 90 west longitude, are in the Central
Time zone and change at 2:00 a.m. local time.

3 Most of Saskatchewan uses Central Standard time all year
round. Areas around Lloydminster are in the Mountain Time zone and
change at 2:00 a.m. local time, as in Alberta.

4 Nunavut covers 3 time zones and all zones change at 02:00 local
time. Southampton Island remains on Eastern Standard time all year
long.

From http://www.statoids.com/tca.html

Uh... See Canada_tz.png in 2005 MEDS archive.

*/

void guess_timezone (double lat, double lng, const Dstr &name,
Dstr &timezone) {
  // Generally following http://www.statoids.com/tca.html
  if (name.contains ("Ontario")) {
    if (lng > -90)
      timezone = ":America/Toronto";
    else
      timezone = ":America/Rainy_River";
  } else if (name.contains ("B.C.")) {
    timezone = ":America/Vancouver";
  } else if (name.contains ("Manitoba")) {
    timezone = ":America/Winnipeg";
  } else if (name.contains ("Nunavut")) {
    // Southampton Island not handled
    if (lng > -85)
      timezone = ":America/Iqaluit";
    else if (lng >= -102)
      timezone = ":America/Rankin_Inlet";
    else
      timezone = ":America/Cambridge_Bay";
  } else if (name.contains ("N.W.T.")) {
    timezone = ":America/Yellowknife";
  } else if (name.contains ("Nova Scotia")) {
    timezone = ":America/Halifax";
  } else if (name.contains ("Qubec")) {
    if (lng > -61.5)
      timezone = "AST4";
    else
      timezone = ":America/Montreal";
  } else if (name.contains ("New Brunswick")) {
    timezone = ":America/Halifax";
  } else if (name.contains ("Newfoundland")) {
    timezone = ":America/St_Johns";
  } else if (name.contains ("Labrador")) {
    // Special case SE tip not handled
    timezone = ":America/Goose_Bay";
  } else if (name.contains ("P.E.I.")) {
    timezone = ":America/Halifax";
  } else {
    assert (0);
  }
}

// STOP YELLING!
void prettify (Dstr &name) {
  Dstr newname;
  bool wascap = 0;
  while (name.length()) {
    char c = name[0];
    name /= 1;
    bool iscap = (c >= 'A' && c <= 'Z');
    if (wascap && iscap)
      c += ' ';
    wascap = iscap;
    newname += c;
  }
  name = newname;
}

/*************************************************************************
The following code for managing timestamps is recycled from Timestamp.cc
in XTide version 2.6.3.  Some modifications have been made.
*************************************************************************/

#define compare_int(a,b) (((int)(a))-((int)(b)))
int
compare_tm (struct tm a, struct tm b) {
  int temp;
  if ((temp = compare_int (a.tm_year, b.tm_year)))
    return temp;
  if ((temp = compare_int (a.tm_mon, b.tm_mon)))
    return temp;
  if ((temp = compare_int (a.tm_mday, b.tm_mday)))
    return temp;
  if ((temp = compare_int (a.tm_hour, b.tm_hour)))
    return temp;
  if ((temp = compare_int (a.tm_min, b.tm_min)))
    return temp;
  return compare_int (a.tm_sec, b.tm_sec);
}

time_t tm2utc (struct tm ht) {
  time_t guess = 0;
  int loopcounter = sizeof(time_t) * 8;
  time_t thebit = ((time_t)1) << (loopcounter-1);

  // Are we signed?
  if (thebit < (time_t)0) {
    switch (loopcounter) {
    case 32:
      // Get this sign thing out of the way.
      if (ht.tm_year + 1900 < 1970)
	guess |= 0x80000000;
      loopcounter = 31;
      break;
    case 64:
      // Had a lot of trouble with overflows.  40 bits should suffice for
      // all reasonable requests.
      if (ht.tm_year + 1900 < 1970)
	guess |= 0xFFFFFF0000000000LL;
      loopcounter = 40;
      break;
    default:
      // Wha?!?
      // Skip the sign bit and hope for the best.
      // You can't just shift thebit right because it propagates the sign bit.
      loopcounter--;
    }
    thebit = ((time_t)1) << (loopcounter-1);
    assert (thebit > (time_t)0);
  } else {
    // We are unsigned.
    // Assuming no overflows, no action required here.
    ;
  }

  int compare;
  struct tm* gm;
  for (; loopcounter; loopcounter--) {
    time_t newguess = guess | thebit;
    // If gmtime fails, we assume that we are too far in the future.
    gm = gmtime(&newguess);
    if (!(compare = (gm ? compare_tm (*gm, ht) : 1)))
      return newguess;
    if (compare < 0)
      guess = newguess;
    assert (thebit > (time_t)0);
    thebit >>= 1;
  }

  assert (!thebit);
  gm = gmtime(&guess);
  assert (!(gm ? compare_tm (*gm, ht) : 1));
  return guess;
}

/*************************************************************************/

int main(int argc, char **argv) {
  if (argc != 2) {
    fprintf (stderr, "\
  Usage:  csv2harmgen medsfile.csv\n\
\n\
  Output:  harmgen_input.txt    (raw format input for harmgen)\n\
           header.txt           (the header file for harmgen)\n");
    exit (-1);
  }

  FILE *fp, *headout, *hgout;
  if (!(fp = fopen (argv[1], "r"))) {
    perror (argv[1]);
    exit (-1);
  }

  // Read csv header
  Dstr name, number, lat, lng, datum, timezone, junk1, junk2;
  name.getline (fp);
  number.getline (fp);
  lat.getline (fp);
  lng.getline (fp);
  datum.getline (fp);
  timezone.getline (fp);
  junk1.getline (fp);
  junk2.getline (fp);
  stripfield (name, "Station_Name");
  stripfield (number, "Station_Number");
  stripfield (lat, "Latitude_Decimal_Degrees");
  stripfield (lng, "Longitude_Decimal_Degrees");
  stripfield (datum, "Datum");
  timezone.trim_tail();
  assert (timezone == "Time_zone,UTC");
  junk1.trim_tail();
  assert (junk1 == "SLEV=Observed Water Level");
  junk2.trim_tail();
  assert (junk2 == "Obs_date,SLEV");

  if (datum == "CD")
    datum = "Chart Datum";
  else if (datum == "IGLD85")
    datum = "International Great Lakes Datum (1985)";
  else if (datum == "IGLD55")
    datum = "International Great Lakes Datum (1955)";
  else if (datum == "")
    datum = "Unknown";
  else {
    fprintf (stderr, "Unhandled datum: '%s'\n", datum.aschar());
    exit (-1);
  }

  double dlat, dlng;
  assert (sscanf (lat.aschar(), "%lf", &dlat) == 1);
  assert (sscanf (lng.aschar(), "%lf", &dlng) == 1);
  dlng = -dlng;

FIXME FIXME

  NEXT TIME AROUND:

  The province and time zone were there in the "catalogs" pages.  Make
  use of this next time.  N.B., they used NT for both N.W.T. and
  Nunavut.

  20050907

FIXME FIXME

  name.repstr (", BC", ", B.C.");
  name.repstr (",B.C", ", B.C.");
  if (name == "YARMOUTH")
    name = "Yarmouth, Nova Scotia";
  else if (name == "Alert")
    name += ", Nunavut";
  else if (name == "Churchill")
    name += ", Manitoba";
  else if (name == "Tuktoyaktuk")
    name += ", N.W.T.";
  else if (name == "SAINT JOHN")
    name = "Saint John, New Brunswick";
  else if (name == "ST JOHNS")
    name = "St. Johns, Newfoundland";
  else if (name == "JOE BATTS ARM")
    name = "Joe Batts Arm, Newfoundland";
  else if (name == "ARGENTIA")
    name = "Argentia, Newfoundland";
  else if (name == "CHARLOTTETOWN")
    name = "Charlottetown, P.E.I.";
  else if (name == "Queen Charlotte City")
    name += ", B.C.";
  else if (name == "HOLMAN")
    name = "Holman, N.W.T.";
  else if (name == "HALIFAX")
    name = "Halifax, Nova Scotia";
  else if (name == "NAIN")
    name = "Nain, Labrador";
  else if (name == "Qikiqtarjuaq")
    name += ", Nunavut";
  else if (name == "PEACE BRIDGE BELOW")
    name = "Peace Bridge Below, Ontario";
  else if (name == "BEDFORD INSTITUTE")
    name = "Bedford Institute, Nova Scotia";
  else if (name == "NORTH SYDNEY")
    name = "North Sydney, Nova Scotia";
  else if (name == "PORT AUX BASQUES")
    name = "Port Aux Basques, Newfoundland";
  else if (name == "WOOD ISLANDS")
    name = "Wood Islands, Nova Scotia";
  else if (name == "SHEDIAC BAY")
    name = "Shediac Bay, New Brunswick";
  else if (name == "LOWER ESCUMINAC")
    name = "Lower Escuminac, New Brunswick";
  else if (name == "BELLEDUNE")
    name = "Belledune, New Brunswick";
  else {
    if (name.contains ("B.C.") || name.contains ("Ontario"))
      ;
    else {
      assert (dlat > 45 && dlat < 51);
      assert (dlng > -74 && dlng < -61);
      name += ", Qubec";
    }
  }
  prettify (name);
  guess_timezone (dlat, dlng, name, timezone);

  if (!(headout = fopen ("header.txt", "w"))) {
    perror ("header.txt");
    exit (-1);
  }
  fprintf (headout, "%s\n%s\nmeters\n", name.aschar(), timezone.aschar());
  fprintf (headout, "\
# The harmonic constants used to perform tide predictions for\n\
# locations in Canada are derived from sea level data made available\n\
# by Marine Environmental Data Services, Fisheries and Oceans Canada,\n\
# for public, non-commercial use.  The predictions are not the same as\n\
# those computed by the Canadian Hydrographic Service, which uses its\n\
# own sets of harmonic constants.\n\
# BEGIN HOT COMMENTS\n\
# country: Canada\n\
# source: Derived from MEDS data with Harmgen 2.2\n\
# restriction: Non-commercial use only\n\
# station_id_context: MEDS\n\
# station_id: %s\n\
# datumkind: %s\n\
# !units: meters\n\
# !longitude: -%s\n\
# !latitude: %s\n", number.aschar(), datum.aschar(), lng.aschar(),
lat.aschar());
  fclose (headout);

  // With UTC0 or :UTC, the timezone assert fails.  So much for mktime
  // (again).
#if 0
  assert (setenv ("TZ", "UTC0", 1) == 0);
  tzset ();
  assert (timezone == 0);
  assert (daylight == 0);
#endif

  if (!(hgout = fopen ("harmgen_input.txt", "w"))) {
    perror ("harmgen_input.txt");
    exit (-1);
  }
  Dstr buf;
  while (!buf.getline(fp).isNull()) {
    struct tm tm;
    char xm;
    double lvl;

    assert (sscanf (buf.aschar(), "%d/%d/%d %d:00:00 %cM,%lf",
      &(tm.tm_mday),
      &(tm.tm_mon),
      &(tm.tm_year),
      &(tm.tm_hour), &xm, &lvl) == 6);
    tm.tm_sec = tm.tm_min = tm.tm_isdst = 0;
    tm.tm_mon--;
    tm.tm_year -= 1900;
    if (xm == 'P') {
      if (tm.tm_hour < 12)
        tm.tm_hour += 12;
    } else
      assert (xm == 'A');
    time_t tt = tm2utc(tm);

    fprintf (hgout, "%ld %f\n", tt, lvl);
  }
  fclose (hgout);
  exit (0);
}
